==> jsonstreamreader-1.1.6.tar.gz: lib/jsontoken.dart <==
import 'dart:async';
import 'dart:math';

import 'state/reader/token/jsontokenstate.dart';

//Token Produces by JsonTokenizer(Parser)
class JsonToken {
  static const int PAIR = 0;
  static const int VALUE = 1;
  static const int KEY = 2;
  static const int OBJECTSEPERATOR = 3;
  static const int ARRAYSEPERATOR = 4;
  static const int OBJECTSTART = 5;
  static const int OBJECTEND = 6;
  static const int ARRAYSTART = 7;
  static const int ARRAYEND = 8;
  static const int VALUEPAIR = 9;
  static const int NULL = 10;
  static var _rng = new Random();
  static Map<int, List<MapEntry<String, dynamic>>> batch =
      new Map<int, List<MapEntry<String, dynamic>>>();
  dynamic value;
  int type;
  int position;
  static Map<int, Map<int, JsonToken>> queue =
      new Map<int, Map<int, JsonToken>>();
  bool closed = false;
  String _key;
  static Map<int, List<JsonToken>> partners = new Map<int, List<JsonToken>>();
  static Map<int, List<JsonToken>> parents = new Map<int, List<JsonToken>>();
  static int pos = 0;
  JsonToken parent;
  JsonToken partner;
  static Map<int, int> _quotes = new Map<int, int>();
  int group;
  int _combine = 0;
  // add these this to state
  int lastIndex;
  JsonToken.empty(
      dynamic statusvalue,
      String statuskey,
      int statustype,
      int statusposition,
      int statuslastindex,
      int statusgroup,
      int statuscombine,
      int statusparent,
      int statuspartner,
      int statusparnterId) {
    this.value = statusvalue;
    this._key = statuskey;
    this.type = statustype;
    this.position = statusposition;
    this.lastIndex = statuslastindex;
    this.group = statusgroup;
    this._combine = statuscombine;
    new Timer(new Duration(microseconds: 0), () {
      if (statusparent != null) {
        this.parent = JsonToken.queue[statusgroup][statusparent];
      }
      if (statuspartner != null) {
        this.partner = JsonToken.queue[statusgroup][statuspartner];
      }
    });
  }
  static initialize(int group) {
    if (batch[group] == null) {
      JsonToken.parents[group] = new List<JsonToken>();
      JsonToken.partners[group] = new List<JsonToken>();
      JsonToken._quotes[group] = 0;
      batch[group] = new List<MapEntry<String, dynamic>>();
      JsonToken.queue[group] = new Map<int, JsonToken>();
    }
  }

  JsonToken(dynamic value, int type,
      {int position, int group, int combine = 0}) {
    this._combine = combine;

    this.value = value;
    this.type = type;
    this.group = group;
    this.position = JsonToken.pos;
    this._key = '/';
    JsonToken.pos++;
    //handle : , inside quotes
    if (this.value == '"') {
      JsonToken._quotes[group]++;
    }
    if ((JsonToken._quotes[group] % 2) == 1 && this.type != JsonToken.NULL) {
      this.type = JsonToken.VALUE;
    }
    JsonToken.queue[group][(position != null ? position : this.position)] =
        this;
    switch (this.type) {
      case JsonToken.OBJECTSTART:
        this.lastIndex = 0;
        //parent
        var parents = JsonToken.parents[group];
        this.parent = parents.length > 0 ? parents.last : null;
        parents.add(this);
        //partner
        JsonToken.partners[group].add(this);
        break;
      case JsonToken.OBJECTEND:
        this.partner = JsonToken.partners[group].removeLast();
        this.partner.closed = true;
        this.closed = true;
        var parents = JsonToken.parents[group];
        parents.removeLast();
        this.parent = parents.length > 0 ? parents.last : null;
        break;
      case JsonToken.ARRAYSTART:
        //path
        this.lastIndex = 0;
        //parent
        var parents = JsonToken.parents[group];
        this.parent = parents.length > 0 ? parents.last : null;
        parents.add(this);
        //partner
        JsonToken.partners[group].add(this);
        break;
      case JsonToken.ARRAYEND:
        this.partner = JsonToken.partners[group].removeLast();
        this.partner.closed = true;
        this.closed = true;
        var parents = JsonToken.parents[group];
        parents.removeLast();
        this.parent = parents.length > 0 ? parents.last : null;
        break;
      case JsonToken.VALUE:
        this.parent = JsonToken.parents[group].last;
        break;
      default:
        break;
    }
  }
  static JsonTokenState getState(int group) {
    //unclosed arrays ids
    int statequotes = JsonToken._quotes[group];
    //
    Map<int, Map<String, dynamic>> statequeue =
        new Map<int, Map<String, dynamic>>();
    List<Map<String, dynamic>> statepartners = List<Map<String, dynamic>>();
    List<Map<String, dynamic>> stateparents = new List<Map<String, dynamic>>();

    JsonToken.queue[group].forEach((int key, JsonToken item) {
      statequeue[key] = {
        'position': item.position,
        'key': item.key,
        'value': item.value,
        'combine': item.combine,
        'group': item.group,
        'type': item.type,
        'parent': item.parent == null ? null : item.parent.position,
        'partner': item.partner == null ? null : item.partner.position,
        'lastIndex': item.lastIndex
      };
    });
    JsonToken.partners[group].forEach((JsonToken item) {
      statepartners.add({
        'position': item.position,
        'key': item.key,
        'value': item.value,
        'combine': item.combine,
        'group': item.group,
        'type': item.type,
        'parent': item.parent == null ? null : item.parent.position,
        'partner': item.partner == null ? null : item.partner.position,
        'lastIndex': item.lastIndex
      });
    });
    JsonToken.parents[group].forEach((JsonToken item) {
      stateparents.add({
        'position': item.position,
        'key': item.key,
        'value': item.value,
        'combine': item.combine,
        'group': item.group,
        'type': item.type,
        'parent': item.parent == null ? null : item.parent.position,
        'partner': item.partner == null ? null : item.partner.position,
        'lastIndex': item.lastIndex
      });
    });
    return new JsonTokenState(
        statequeue, statepartners, stateparents, statequotes);
  }

  static void fromState(int groupnumber, JsonTokenState state) {
    //set JsonToken.pos to last position+1

    if (batch[groupnumber] == null) {
      batch[groupnumber] = new List<MapEntry<String, dynamic>>();
      JsonToken.queue[groupnumber] = new Map<int, JsonToken>();
    } else {
      batch[groupnumber] = new List<MapEntry<String, dynamic>>();
      JsonToken.queue[groupnumber] = new Map<int, JsonToken>();
    }
    Map<String, dynamic> data = state.data();
    JsonToken._quotes[groupnumber] = data['quotes'];
    //
    data['queue'].forEach((int key, Map<String, dynamic> item) {
      JsonToken.queue[groupnumber][key] = new JsonToken.empty(
          item['value'],
          item['key'],
          item['type'],
          item['position'],
          item['lastIndex'],
          item['group'],
          item['combine'],
          item['parent'],
          item['partner'],
          item['partnerId']);
      JsonToken.pos = item['position'] + 2;
    });
    JsonToken.parents[groupnumber] = new List();
    data['parents'].forEach((Map<String, dynamic> item) {
      JsonToken.parents[groupnumber].add(new JsonToken.empty(
          item['value'],
          item['key'],
          item['type'],
          item['position'],
          item['lastIndex'],
          item['group'],
          item['combine'],
          item['parent'],
          item['partner'],
          item['partnerId']));
    });
    JsonToken.partners[groupnumber] = new List();
    data['partners'].forEach((Map<String, dynamic> item) {
      JsonToken.partners[groupnumber].add(new JsonToken.empty(
          item['value'],
          item['key'],
          item['type'],
          item['position'],
          item['lastIndex'],
          item['group'],
          item['combine'],
          item['parent'],
          item['partner'],
          item['partnerId']));
    });
  }

  JsonToken append(JsonToken token) {
    if (this.value.toString().endsWith("\\") && token.value == "\"") {
      JsonToken._quotes[group]--;
      this.value =
          this.value.toString().replaceAll(new RegExp(r"\\$"), "") + '\\"';
      this._key = token._key;
      token.value = null;
      token.type = JsonToken.NULL;
      return this;
    } else if (this.type == JsonToken.VALUE && token.type == JsonToken.VALUE) {
      var r = token;
      this.value += r.value;
      this.value = this.value;
      token.value = null;
      token.type = JsonToken.NULL;
      return this;
    } else if (this.type == JsonToken.VALUEPAIR &&
        token.type == JsonToken.VALUE) {
      var r = token;
      this.value += r.value;
      this.value = this.value;
      token.value = null;
      token.type = JsonToken.NULL;
      return this;
    } else if (this.type == JsonToken.VALUE &&
        token.type == JsonToken.OBJECTSEPERATOR) {
      this.type = JsonToken.KEY;
      token._key = this.parent._key +
          this.value.replaceAll("\"", "").replaceAll("'", "").trim() +
          "/";
      return token;
    } else if (this.type == JsonToken.OBJECTSEPERATOR &&
        token.type == JsonToken.VALUE) {
      token.type = JsonToken.VALUEPAIR;
      token._key = this._key;
      return token;
    } else if (this.type == JsonToken.VALUEPAIR &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      token._key = this.parent._key;
      //release token
      List<String> pieces = this.key.split('/');
      JsonToken.stream(
          this.group, token._key, {pieces[pieces.length - 2]: this.value});
      return token;
    } else if (this.type == JsonToken.VALUEPAIR &&
        token.type == JsonToken.OBJECTEND) {
      token._key = this.parent.key;
      //release token
      List<String> pieces = this.key.split('/');
      JsonToken.stream(
          this.group, token._key, {pieces[pieces.length - 2]: this.value});
      return token;
    } else if (this.type == JsonToken.VALUE &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      //release token
      String newKey = this.parent._key + this.parent.lastIndex.toString() + '/';
      if (this._combine == 1) {
        JsonToken.stream(this.group, this.parent.key,
            {this.parent.lastIndex.toString(): this.value});
      } else if (this._combine == 2) {
        String subkey = '';
        List<String> pieces = this.parent.key.split('/');
        pieces.removeWhere((String piece) {
          return piece == null || piece == '';
        });
        if (pieces.length > 1) {
          pieces.getRange(1, pieces.length).toList().forEach((String k) {
            subkey += k + '/';
          });
        }
        JsonToken.stream(this.group, pieces[0], {
          '/' + subkey + this.parent.lastIndex.toString() + '/': this.value
        });
      } else {
        JsonToken.stream(this.group, newKey, this.value);
      }
      this._key = newKey;
      this.parent.lastIndex++;
      return token;
    } else if (this.type == JsonToken.VALUE &&
        token.type == JsonToken.ARRAYEND) {
      String newKey = this.parent._key + this.parent.lastIndex.toString() + '/';
      //release token
      if (this._combine == 1) {
        JsonToken.stream(this.group, this.parent._key,
            {this.parent.lastIndex.toString(): this.value});
      } else if (this._combine == 2) {
        String subkey = '';
        List<String> pieces = this.parent.key.split('/');
        pieces.removeWhere((String piece) {
          return piece == null || piece == '';
        });
        if (pieces.length > 1) {
          pieces.getRange(1, pieces.length).toList().forEach((String k) {
            subkey += k + '/';
          });
        }
        JsonToken.stream(this.group, pieces[0], {
          '/' + subkey + this.parent.lastIndex.toString() + '/': this.value
        });
      } else {
        JsonToken.stream(this.group, newKey, this.value);
      }
      this._key = newKey;
      this.parent.lastIndex++;
      return token;
    } else if (this.type == JsonToken.ARRAYSTART &&
        token.type == JsonToken.OBJECTSTART) {
      token._key = this._key + token.parent.lastIndex.toString() + '/';
      return token;
    } else if (this.type == JsonToken.OBJECTEND &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      token._key = this.parent._key;
      this.parent.lastIndex++;
      return token;
    } else if (this.type == JsonToken.ARRAYSEPERATOR &&
        token.type == JsonToken.OBJECTSTART) {
      token._key = this._key + token.parent.lastIndex.toString() + '/';
      return token;
    } else if (this.type == JsonToken.OBJECTEND &&
        token.type == JsonToken.ARRAYEND) {
      token._key = this.parent._key;
      return token;
    } else if (this.type == JsonToken.ARRAYSTART &&
        token.type == JsonToken.ARRAYSTART) {
      token._key = this._key + this.parent.lastIndex.toString() + '/';
      this.parent.lastIndex++;
      return token;
    } else if (this.type == JsonToken.ARRAYEND &&
        token.type == JsonToken.ARRAYEND) {
      token._key = partner._key;
      return token;
    } else if (this.type == JsonToken.ARRAYEND &&
        token.type == JsonToken.OBJECTEND) {
      token._key = partner._key;
      return token;
    } else {
      token._key = this._key;
      return token;
    }
  }

  String get key {
    return this._key;
  }

  int get combine {
    return this._combine;
  }

  toString() {
    return this.value;
  }

  static int getType(String value) {
    if (value == null) {
      return JsonToken.NULL;
    } else if (value == ']') {
      return JsonToken.ARRAYEND;
    } else if (value == '[') {
      return JsonToken.ARRAYSTART;
    } else if (value == '{') {
      return JsonToken.OBJECTSTART;
    } else if (value == '}') {
      return JsonToken.OBJECTEND;
    } else if (value == ',') {
      return JsonToken.ARRAYSEPERATOR;
    } else if (value == ':') {
      return JsonToken.OBJECTSEPERATOR;
    } else {
      return JsonToken.VALUE;
    }
  }

  static stream(int group, String key, dynamic value) {
    if (batch[group].length > 0) {
      if (key == batch[group].last.key) {
        if (batch[group]
                .last
                .value
                .runtimeType
                .toString()
                .contains("HashMap") &&
            value.runtimeType.toString().contains("HashMap")) {
          batch[group].last.value.addAll(value);
        } else {
          batch[group].add(new MapEntry(key, value));
        }
      } else {
        batch[group].add(new MapEntry(key, value));
      }
    } else {
      batch[group].add(new MapEntry(key, value));
    }
  }

  static identity() {
    return _rng.nextInt(100000);
  }
}

==> jsonstreamreader-1.1.8.tar.gz: lib/jsontoken.dart <==
import 'dart:async';
import 'dart:math';

import 'state/reader/token/jsontokenstate.dart';

//Token Produces by JsonTokenizer(Parser)
class JsonToken {
  static const int PAIR = 0;
  static const int VALUE = 1;
  static const int KEY = 2;
  static const int OBJECTSEPERATOR = 3;
  static const int ARRAYSEPERATOR = 4;
  static const int OBJECTSTART = 5;
  static const int OBJECTEND = 6;
  static const int ARRAYSTART = 7;
  static const int ARRAYEND = 8;
  static const int VALUEPAIR = 9;
  static const int NULL = 10;
  static var _rng = new Random();
  static Map<int, List<MapEntry<String, dynamic>>> batch =
      new Map<int, List<MapEntry<String, dynamic>>>();
  dynamic value;
  int type;
  int position;
  static Map<int, Map<int, JsonToken>> queue =
      new Map<int, Map<int, JsonToken>>();
  bool closed = false;
  String _key;
  static Map<int, List<JsonToken>> partners = new Map<int, List<JsonToken>>();
  static Map<int, List<JsonToken>> parents = new Map<int, List<JsonToken>>();
  static int pos = 0;
  JsonToken parent;
  JsonToken partner;
  static Map<int, int> _quotes = new Map<int, int>();
  int group;
  int _combine = 0;
  // add these this to state
  int lastIndex;
  JsonToken.empty(
      dynamic statusvalue,
      String statuskey,
      int statustype,
      int statusposition,
      int statuslastindex,
      int statusgroup,
      int statuscombine,
      int statusparent,
      int statuspartner,
      int statusparnterId) {
    this.value = statusvalue;
    this._key = statuskey;
    this.type = statustype;
    this.position = statusposition;
    this.lastIndex = statuslastindex;
    this.group = statusgroup;
    this._combine = statuscombine;
    new Timer(new Duration(microseconds: 0), () {
      if (statusparent != null) {
        this.parent = JsonToken.queue[statusgroup][statusparent];
      }
      if (statuspartner != null) {
        this.partner = JsonToken.queue[statusgroup][statuspartner];
      }
    });
  }
  static initialize(int group) {
    if (batch[group] == null) {
      JsonToken.parents[group] = new List<JsonToken>();
      JsonToken.partners[group] = new List<JsonToken>();
      JsonToken._quotes[group] = 0;
      batch[group] = new List<MapEntry<String, dynamic>>();
      JsonToken.queue[group] = new Map<int, JsonToken>();
    }
  }

  JsonToken(dynamic value, int type,
      {int position, int group, int combine = 0}) {
    this._combine = combine;

    this.value = value;
    this.type = type;
    this.group = group;
    this.position = JsonToken.pos;
    this._key = '/';
    JsonToken.pos++;
    //handle : , inside quotes
    if (this.value == '"') {
      JsonToken._quotes[group]++;
    }
    if ((JsonToken._quotes[group] % 2) == 1 && this.type != JsonToken.NULL) {
      this.type = JsonToken.VALUE;
    }
    JsonToken.queue[group][(position != null ? position : this.position)] =
        this;
    switch (this.type) {
      case JsonToken.OBJECTSTART:
        this.lastIndex = 0;
        //parent
        var parents = JsonToken.parents[group];
        this.parent = parents.length > 0 ? parents.last : null;
        parents.add(this);
        //partner
        JsonToken.partners[group].add(this);
        break;
      case JsonToken.OBJECTEND:
        this.partner = JsonToken.partners[group].removeLast();
        this.partner.closed = true;
        this.closed = true;
        var parents = JsonToken.parents[group];
        parents.removeLast();
        this.parent = parents.length > 0 ? parents.last : null;
        break;
      case JsonToken.ARRAYSTART:
        //path
        this.lastIndex = 0;
        //parent
        var parents = JsonToken.parents[group];
        this.parent = parents.length > 0 ? parents.last : null;
        parents.add(this);
        //partner
        JsonToken.partners[group].add(this);
        break;
      case JsonToken.ARRAYEND:
        this.partner = JsonToken.partners[group].removeLast();
        this.partner.closed = true;
        this.closed = true;
        var parents = JsonToken.parents[group];
        parents.removeLast();
        this.parent = parents.length > 0 ? parents.last : null;
        break;
      case JsonToken.VALUE:
        this.parent = JsonToken.parents[group].last;
        break;
      default:
        break;
    }
  }
  static JsonTokenState getState(int group) {
    //unclosed arrays ids
    int statequotes = JsonToken._quotes[group];
    //
    Map<int, Map<String, dynamic>> statequeue =
        new Map<int, Map<String, dynamic>>();
    List<Map<String, dynamic>> statepartners = new List<Map<String, dynamic>>();
    List<Map<String, dynamic>> stateparents = new List<Map<String, dynamic>>();

    JsonToken.queue[group].forEach((int key, JsonToken item) {
      statequeue[key] = {
        'position': item.position,
        'key': item.key,
        'value': item.value,
        'combine': item.combine,
        'group': item.group,
        'type': item.type,
        'parent': item.parent == null ? null : item.parent.position,
        'partner': item.partner == null ? null : item.partner.position,
        'lastIndex': item.lastIndex
      };
    });
    JsonToken.partners[group].forEach((JsonToken item) {
      statepartners.add({
        'position': item.position,
        'key': item.key,
        'value': item.value,
        'combine': item.combine,
        'group': item.group,
        'type': item.type,
        'parent': item.parent == null ? null : item.parent.position,
        'partner': item.partner == null ? null : item.partner.position,
        'lastIndex': item.lastIndex
      });
    });
    JsonToken.parents[group].forEach((JsonToken item) {
      stateparents.add({
        'position': item.position,
        'key': item.key,
        'value': item.value,
        'combine': item.combine,
        'group': item.group,
        'type': item.type,
        'parent': item.parent == null ? null : item.parent.position,
        'partner': item.partner == null ? null : item.partner.position,
        'lastIndex': item.lastIndex
      });
    });
    return new JsonTokenState(
        statequeue, statepartners, stateparents, statequotes);
  }

  static void fromState(int groupnumber, JsonTokenState state) {
    //set JsonToken.pos to last position+1

    if (batch[groupnumber] == null) {
      batch[groupnumber] = new List<MapEntry<String, dynamic>>();
      JsonToken.queue[groupnumber] = new Map<int, JsonToken>();
    } else {
      batch[groupnumber] = new List<MapEntry<String, dynamic>>();
      JsonToken.queue[groupnumber] = new Map<int, JsonToken>();
    }
    Map<String, dynamic> data = state.data();
    JsonToken._quotes[groupnumber] = data['quotes'];
    //
    data['queue'].forEach((int key, Map<String, dynamic> item) {
      JsonToken.queue[groupnumber][key] = new JsonToken.empty(
          item['value'],
          item['key'],
          item['type'],
          item['position'],
          item['lastIndex'],
          item['group'],
          item['combine'],
          item['parent'],
          item['partner'],
          item['partnerId']);
      JsonToken.pos = item['position'] + 2;
    });
    JsonToken.parents[groupnumber] = new List();
    data['parents'].forEach((Map<String, dynamic> item) {
      JsonToken.parents[groupnumber].add(new JsonToken.empty(
          item['value'],
          item['key'],
          item['type'],
          item['position'],
          item['lastIndex'],
          item['group'],
          item['combine'],
          item['parent'],
          item['partner'],
          item['partnerId']));
    });
    JsonToken.partners[groupnumber] = new List();
    data['partners'].forEach((Map<String, dynamic> item) {
      JsonToken.partners[groupnumber].add(new JsonToken.empty(
          item['value'],
          item['key'],
          item['type'],
          item['position'],
          item['lastIndex'],
          item['group'],
          item['combine'],
          item['parent'],
          item['partner'],
          item['partnerId']));
    });
  }

  JsonToken append(JsonToken token) {
    if (this.value.toString().endsWith("\\") && token.value == "\"") {
      JsonToken._quotes[group]--;
      this.value =
          this.value.toString().replaceAll(new RegExp(r"\\$"), "") + '\\"';
      this._key = token._key;
      token.value = null;
      token.type = JsonToken.NULL;
      return this;
    } else if (this.type == JsonToken.VALUE && token.type == JsonToken.VALUE) {
      var r = token;
      this.value += r.value;
      this.value = this.value;
      token.value = null;
      token.type = JsonToken.NULL;
      return this;
    } else if (this.type == JsonToken.VALUEPAIR &&
        token.type == JsonToken.VALUE) {
      var r = token;
      this.value += r.value;
      this.value = this.value;
      token.value = null;
      token.type = JsonToken.NULL;
      return this;
    } else if (this.type == JsonToken.VALUE &&
        token.type == JsonToken.OBJECTSEPERATOR) {
      this.type = JsonToken.KEY;
      token._key = this.parent._key +
          this.value.replaceAll("\"", "").replaceAll("'", "").trim() +
          "/";
      return token;
    } else if (this.type == JsonToken.OBJECTSEPERATOR &&
        token.type == JsonToken.VALUE) {
      token.type = JsonToken.VALUEPAIR;
      token._key = this._key;
      return token;
    } else if (this.type == JsonToken.VALUEPAIR &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      token._key = this.parent._key;
      //release token
      List<String> pieces = this.key.split('/');
      JsonToken.stream(
          this.group, token._key, {pieces[pieces.length - 2]: this.value});
      return token;
    } else if (this.type == JsonToken.VALUEPAIR &&
        token.type == JsonToken.OBJECTEND) {
      token._key = this.parent.key;
      //release token
      List<String> pieces = this.key.split('/');
      JsonToken.stream(
          this.group, token._key, {pieces[pieces.length - 2]: this.value});
      return token;
    } else if (this.type == JsonToken.VALUE &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      //release token
      String newKey = this.parent._key + this.parent.lastIndex.toString() + '/';
      if (this._combine == 1) {
        JsonToken.stream(this.group, this.parent.key,
            {this.parent.lastIndex.toString(): this.value});
      } else if (this._combine == 2) {
        String subkey = '';
        List<String> pieces = this.parent.key.split('/');
        pieces.removeWhere((String piece) {
          return piece == null || piece == '';
        });
        String start;
        if (pieces.length > 1) {
          pieces.getRange(1, pieces.length).toList().forEach((String k) {
            subkey += k + '/';
          });
          start = pieces[0];
        } else {
          start = token._key;
        }
        JsonToken.stream(this.group, start, {
          '/' + subkey + this.parent.lastIndex.toString() + '/': this.value
        });
      } else {
        JsonToken.stream(this.group, newKey, this.value);
      }
      this._key = newKey;
      this.parent.lastIndex++;
      return token;
    } else if (this.type == JsonToken.VALUE &&
        token.type == JsonToken.ARRAYEND) {
      String newKey = this.parent._key + this.parent.lastIndex.toString() + '/';
      //release token
      if (this._combine == 1) {
        JsonToken.stream(this.group, this.parent._key,
            {this.parent.lastIndex.toString(): this.value});
      } else if (this._combine == 2) {
        String subkey = '';
        List<String> pieces = this.parent.key.split('/');
        pieces.removeWhere((String piece) {
          return piece == null || piece == '';
        });
        String start;
        if (pieces.length > 1) {
          pieces.getRange(1, pieces.length).toList().forEach((String k) {
            subkey += k + '/';
          });
          start = pieces[0];
        } else {
          start = token._key;
        }
        JsonToken.stream(this.group, start, {
          '/' + subkey + this.parent.lastIndex.toString() + '/': this.value
        });
      } else {
        JsonToken.stream(this.group, newKey, this.value);
      }
      this._key = newKey;
      this.parent.lastIndex++;
      return token;
    } else if (this.type == JsonToken.ARRAYSTART &&
        token.type == JsonToken.OBJECTSTART) {
      token._key = this._key + token.parent.lastIndex.toString() + '/';
      return token;
    } else if (this.type == JsonToken.OBJECTEND &&
        token.type == JsonToken.ARRAYSEPERATOR) {
      token._key = this.parent._key;
      this.parent.lastIndex++;
      return token;
    } else if (this.type == JsonToken.ARRAYSEPERATOR &&
        token.type == JsonToken.OBJECTSTART) {
      token._key = this._key + token.parent.lastIndex.toString() + '/';
      return token;
    } else if (this.type == JsonToken.OBJECTEND &&
        token.type == JsonToken.ARRAYEND) {
      token._key = this.parent._key;
      return token;
    } else if (this.type == JsonToken.ARRAYSTART &&
        token.type == JsonToken.ARRAYSTART) {
      token._key = this._key + this.parent.lastIndex.toString() + '/';
      this.parent.lastIndex++;
      return token;
    } else if (this.type == JsonToken.ARRAYEND &&
        token.type == JsonToken.ARRAYEND) {
      token._key = partner._key;
      return token;
    } else if (this.type == JsonToken.ARRAYEND &&
        token.type == JsonToken.OBJECTEND) {
      token._key = partner._key;
      return token;
    } else {
      token._key = this._key;
      return token;
    }
  }

  String get key {
    return this._key;
  }

  int get combine {
    return this._combine;
  }

  toString() {
    return this.value;
  }

  static int getType(String value) {
    if (value == null) {
      return JsonToken.NULL;
    } else if (value == ']') {
      return JsonToken.ARRAYEND;
    } else if (value == '[') {
      return JsonToken.ARRAYSTART;
    } else if (value == '{') {
      return JsonToken.OBJECTSTART;
    } else if (value == '}') {
      return JsonToken.OBJECTEND;
    } else if (value == ',') {
      return JsonToken.ARRAYSEPERATOR;
    } else if (value == ':') {
      return JsonToken.OBJECTSEPERATOR;
    } else {
      return JsonToken.VALUE;
    }
  }

  static stream(int group, String key, dynamic value) {
    if (batch[group].length > 0) {
      if (key == batch[group].last.key) {
        if (batch[group]
                .last
                .value
                .runtimeType
                .toString()
                .contains("HashMap") &&
            value.runtimeType.toString().contains("HashMap")) {
          batch[group].last.value.addAll(value);
        } else {
          batch[group].add(new MapEntry(key, value));
        }
      } else {
        batch[group].add(new MapEntry(key, value));
      }
    } else {
      batch[group].add(new MapEntry(key, value));
    }
  }

  static identity() {
    return _rng.nextInt(100000);
  }
}
